"""
PeeringDB API
"""
import logging, sys
from importlib import import_module
import pkg_resources

__version__ = pkg_resources.require("peeringdb")[0].version
_log_level = logging.INFO


def _config_logs(lvl=None, name=None):
    """
    Set up or change logging configuration.

    _config_logs() => idempotent setup;
    _config_logs(L) => change log level
    """
    # print('_config_log', 'from %s' %name if name else '')
    FORMAT = "%(message)s"
    # maybe better for log files
    # FORMAT='[%(levelname)s]:%(message)s',

    # Reset handlers
    for h in list(logging.root.handlers):
        logging.root.removeHandler(h)

    global _log_level
    if lvl:
        _log_level = lvl

    logging.basicConfig(level=_log_level, format=FORMAT, stream=sys.stdout)
    _log = logging.getLogger(__name__)
    _log.setLevel(_log_level)

    # external
    for log in ["urllib3", "asyncio"]:
        logging.getLogger(log).setLevel(_log_level)


class BackendError(Exception):
    pass


# Map external module names to adaptor modules
SUPPORTED_BACKENDS = {
    "django_peeringdb": "django_peeringdb.client_adaptor",
}

__backend = None


def backend_initialized():
    return __backend is not None


def _get():
    global __backend
    if __backend:
        return __backend
    else:
        raise BackendError("Backend not initialized")


def get_backend():
    return _get()[0]


def get_backend_info():
    return _get()[1]


def initialize_backend(name, **kwargs):
    global __backend
    if __backend:
        raise BackendError("Backend already initialized")

    try:
        modname = SUPPORTED_BACKENDS[name]
    except KeyError:
        raise ValueError("Not a supported backend module: '{}'".format(name))
    # Load internal module associated with the ORM module
    supportmod = import_module(modname)
    # Backend is any object returned from load_backend
    B = supportmod.load_backend(**kwargs)

    B.Backend.setup()
    __backend = (B.Backend(), (name, B.__version__))


# TODO
# def is_valid_backend(backend): ...

# namespace imports
from peeringdb.client import Client as PeeringDB
from peeringdb.config import load_config
from peeringdb.resource import get_resource



import logging, re
from functools import reduce
import json

from django.core import serializers

from peeringdb import resource, get_backend


def split_ref(string):
    """ splits a string into (tag, id) """
    re_tag = re.compile(r"^(?P<tag>[a-zA-Z]+)[\s-]*(?P<pk>\d+)$")
    m = re_tag.search(string)
    if not m:
        raise ValueError("unable to split string '{}'".format(string))

    return (m.group("tag").lower(), int(m.group("pk")))


def pretty_speed(value):
    if not value:
        return ""
    try:
        value = int(value)
        if value >= 1000000:
            return "%dT" % (value / 10 ** 6)
        elif value >= 1000:
            return "%dG" % (value / 10 ** 3)
        else:
            return "%dM" % value
    except ValueError:
        return value


def prompt(msg, default=None):
    "Prompt for input"
    if default is not None:
        msg = "{} ({})".format(msg, repr(default))
    msg = "{}: ".format(msg)
    try:
        s = input(msg)
    except KeyboardInterrupt:
        exit(1)
    except EOFError:
        s = ""
    if not s:
        s = default
    return s


def group_fields(concrete):
    "Partition a concrete's fields into groups based on type"
    GROUPS = ("scalars", "single_refs", "many_refs")
    B = get_backend()
    ret = {kind: {} for kind in GROUPS}
    fields = B.get_fields(concrete)
    for field in fields:
        name = field.name
        related, multiple = B.is_field_related(concrete, name)

        if related:
            if multiple:
                group = ret["many_refs"]
            else:
                group = ret["single_refs"]
        else:
            group = ret["scalars"]

        group[name] = field
    return ret


def limit_mem(limit=(4 * 1024 ** 3)):
    "Set soft memory limit"
    rsrc = resource.RLIMIT_DATA
    soft, hard = resource.getrlimit(rsrc)
    resource.setrlimit(rsrc, (limit, hard))  # 4GB
    softnew, _ = resource.getrlimit(rsrc)
    assert softnew == limit

    _log = logging.getLogger(__name__)
    _log.debug("Set soft memory limit: %s => %s", soft, softnew)


def client_dump(client, path):
    "Serialize all objects into JSON files in directory"
    assert path.is_dir(), path
    for q in client.tags.all():
        ser = serializers.serialize("json", q.all())
        outpath = path / "{}.json".format(q.res.tag)
        with open(outpath, "w") as out:
            print("Writing {}".format(outpath))
            j = json.loads(ser)
            json.dump(j, out, indent=4, sort_keys=True)
            # out.write(ser)


def client_load(client, path):
    "Deserialize from JSON files under directory"
    for q in client.tags.all():
        respath = path / "{}.json".format(q.res.tag)
        with open(str(respath)) as fin:
            des = serializers.deserialize("json", fin)
            for obj in des:
                obj.save()



"""
RPC / REST client implementation module
"""
import re
import requests
from twentyc.rpc import RestClient
from twentyc.rpc.client import (
    NotFoundException,
    PermissionDeniedException,
    InvalidRequestException,
)

import peeringdb


class CompatibilityError(Exception):
    pass


class Fetcher(RestClient):
    """
    REST client with some patches
    """

    def __init__(self, **kwargs):
        # self.return_error = True
        super().__init__(**kwargs)

    def _req(self, func):
        try:
            return func(), None
        except NotFoundException as e:  # 404
            return {}, e
        except PermissionDeniedException as e:  # 401
            return {}, e
        except InvalidRequestException as e:
            pattern = "client version is incompatible"
            error = e.extra["meta"]["error"]
            if re.search(pattern, error):
                raise CompatibilityError(error)
            raise

    def fetch(self, R, pk, depth):
        return self._req(lambda: self.get(R.tag, pk, depth=depth))

    def fetch_latest(self, R, pk, depth, since=None):
        backend = peeringdb.get_backend()
        if since is None:
            since = backend.last_change(backend.get_concrete(R))
        if since:
            since = since + 1
        return self._req(lambda: self.get(R.tag, pk, since=since, depth=depth))

    def fetch_all(self, R, depth, params={}):
        params = {
            k: ",".join(map(str, v)) if isinstance(v, (list, tuple)) else v
            for k, v in params.items()
        }
        return self._req(lambda: self.all(R.tag, depth=depth, **params))

    def fetch_all_latest(self, R, depth, params={}, since=None):
        backend = peeringdb.get_backend()

        if since is None:
            since = backend.last_change(backend.get_concrete(R))

        if since:
            since = since + 1
        params = {
            k: ",".join(map(str, v)) if isinstance(v, (list, tuple)) else v
            for k, v in params.items()
        }
        return self._req(lambda: self.all(R.tag, since=since, depth=depth, **params))

    def fetch_deleted(self, R, pk, depth):
        def req():
            return self.all(R.tag, id=pk, since=1, depth=depth)

        return self._req(req)

    # fixme:
    # RestClient monkeypatches to add headers to request

    def get(self, typ, id, **kwargs):
        """
        Load type by id
        """
        return self._load(self._request(typ, id=id, params=kwargs))

    def _request(self, typ, id=0, method="GET", params=None, data=None, url=None):
        """
        send the request, return response obj
        """
        backend, backend_version = peeringdb.get_backend_info()
        user_agent = "PeeringDB/{} {}/{}".format(
            peeringdb.__version__, backend, backend_version
        )
        headers = {
            "Accept": "application/json",
            "User-Agent": user_agent,
        }
        auth = None

        if self.user:
            auth = (self.user, self.password)
        if not url:
            if id:
                url = "{}/{}/{}".format(self.url, typ, id)
            else:
                url = "{}/{}".format(self.url, typ)

        return requests.request(
            method, url, params=params, data=data, auth=auth, headers=headers
        )



"""
PeeringDB configuration module.

This defines config schemas and related I/O.
"""
import logging
import os

import munge
from munge.util import recursive_update

from confu import schema as _schema, generator

from peeringdb.util import prompt

DEFAULT_CONFIG_DIR = "~/.peeringdb"


class ClientSchema(_schema.Schema):
    """
    Default confu schema for PeeringDB client.
    """

    class SyncSchema(_schema.Schema):
        url = _schema.Url("url", default="https://www.peeringdb.com/api")
        user = _schema.Str("user", blank=True, default="")
        password = _schema.Str("password", blank=True, default="")
        strip_tz = _schema.Int("strip_tz", default=1)  # FIXME no boolean?
        only = _schema.List("only", item=_schema.Str(), default=[])
        timeout = _schema.Int("timeout", default=0)

    class OrmSchema(_schema.Schema):
        class OrmDbSchema(_schema.Schema):
            engine = _schema.Str("engine", default="sqlite3")
            name = _schema.Str("name", default="peeringdb.sqlite3")
            host = _schema.Str("host", blank=True, default="")
            port = _schema.Int("port", default=0)
            user = _schema.Str("user", blank=True, default="")
            password = _schema.Str("password", blank=True, default="")

        secret_key = _schema.Str("secret_key", blank=True, default="")
        backend = _schema.Str("backend", default="django_peeringdb")
        migrate = _schema.Bool("migrate", default=True)
        database = OrmDbSchema()

    sync = SyncSchema()
    orm = OrmSchema()


CLIENT_SCHEMA = ClientSchema()


def default_config(schema=CLIENT_SCHEMA):
    "Get default config values."
    return generator.generate(schema)
    # return confu.config.Config(schema, data=generator.generate(schemas))


def read_config(conf_dir=DEFAULT_CONFIG_DIR):
    "Find and read config file for a directory, return None if not found."

    conf_path = os.path.expanduser(conf_dir)
    if not os.path.exists(conf_path):
        # only throw if not default
        if conf_dir != DEFAULT_CONFIG_DIR:
            raise OSError("Config directory not found at {}".format(conf_path))

    return munge.load_datafile("config", conf_path, default=None)


def load_config(conf_dir=DEFAULT_CONFIG_DIR, schema=CLIENT_SCHEMA):
    """
    Load config files from the specified directory, using defaults for missing values.
    Directory should contain a file named config.<ext> where <ext> is a
    supported config file format.
    """
    data = default_config(schema)
    config = read_config(conf_dir)
    if config:
        recursive_update(data, config)
    return data


class _OldClientSchema(_schema.Schema):
    class PeeringDBSchema(_schema.Schema):
        url = _schema.Url("url", default="https://www.peeringdb.com/api")
        user = _schema.Str("user", blank=True, default="")
        password = _schema.Str("password", blank=True, default="")
        timeout = _schema.Int("timeout", default=0)

    class DatabaseSchema(_schema.Schema):
        engine = _schema.Str("engine", default="sqlite3")
        name = _schema.Str("name", default="peeringdb.sqlite3")
        host = _schema.Str("host", blank=True, default="")
        port = _schema.Int("port", default=0)
        user = _schema.Str("user", blank=True, default="")
        password = _schema.Str("password", blank=True, default="")

    __config_dir__ = _schema.Str("__config_dir__", blank=True, default="")
    peeringdb = PeeringDBSchema()
    database = DatabaseSchema()


_OLD_SCHEMA = _OldClientSchema()


def detect_old(data):
    "Check for a config file with old schema"
    if not data:
        return False
    ok, errors, warnings = _schema.validate(_OLD_SCHEMA, data)
    return ok and not (errors or warnings)


def convert_old(data):
    "Convert config data with old schema to new schema"
    ret = default_config()
    ret["sync"].update(data.get("peeringdb", {}))
    ret["orm"]["database"].update(data.get("database", {}))
    return ret


def write_config(
    data, conf_dir=DEFAULT_CONFIG_DIR, codec="yaml", backup_existing=False
):
    """
    Write config values to a file.

    Arguments:
        - conf_dir<str>: path to output directory
        - codec<str>: output field format
        - backup_existing<bool>: if a config file exists,
            make a copy before overwriting
    """
    if not codec:
        codec = "yaml"
    codec = munge.get_codec(codec)()
    conf_dir = os.path.expanduser(conf_dir)
    if not os.path.exists(conf_dir):
        os.mkdir(conf_dir)

    # Check for existing file, back up if necessary
    outpath = os.path.join(conf_dir, "config." + codec.extensions[0])
    if backup_existing and os.path.exists(outpath):
        os.rename(outpath, outpath + ".bak")
    codec.dump(data, open(outpath, "w"))


def prompt_config(sch, defaults=None, path=None):
    """
    Utility function to recursively prompt for config values

    Arguments:
        - defaults<dict>: default values used for empty inputs
        - path<str>: path to prepend to config keys (eg. "path.keyname")
    """
    out = {}
    for name, attr in sch.attributes():
        fullpath = name
        if path:
            fullpath = "{}.{}".format(path, name)
        if defaults is None:
            defaults = {}
        default = defaults.get(name)
        if isinstance(attr, _schema.Schema):
            # recurse on sub-schema
            value = prompt_config(attr, defaults=default, path=fullpath)
        else:
            if default is None:
                default = attr.default
            if default is None:
                default = ""
            value = prompt(fullpath, default)
        out[name] = value

    return sch.validate(out)


# def old_to_new(old_config):
#     "Migrate an old-schema config to new schema"
#     _OLD_SCHEMA.validate(old_config)



"""
Wrapper module for asynchronous task handling.
"""
import asyncio, inspect


class UpdateTask(asyncio.Task):
    """Wrap a coroutine in a future"""

    def __init__(self, coro, desc, loop=None):
        """
        Arguments:
            - coro: awaitable object
            - desc<tuple>: (Resource, key)
        """
        assert asyncio.iscoroutine(coro), coro
        self._desc = desc
        super().__init__(coro)

    def __repr__(self):
        res, pk = self._desc
        return "<UpdateTask: ({}, {})>".format(res.tag, pk)


def gather(jobs):
    "Aggregate and collect jobs"
    return asyncio.gather(*jobs, return_exceptions=False)


def wrap_generator(func):
    """
    Decorator to convert a generator function to an async function which collects
    and returns generator results, returning a list if there are multiple results
    """

    async def _wrapped(*a, **k):
        r, ret = None, []
        gen = func(*a, **k)
        while True:
            try:
                item = gen.send(r)
            except StopIteration:
                break
            if inspect.isawaitable(item):
                r = await item
            else:
                r = item
            ret.append(r)

        if len(ret) == 1:
            return ret.pop()
        return ret

    return _wrapped


def run_task(func):
    """
    Decorator to wrap an async function in an event loop.
    Use for main sync interface methods.
    """

    def _wrapped(*a, **k):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(func(*a, **k))

    return _wrapped



import collections
import sys
from peeringdb.util import pretty_speed


class WhoisFormat:
    def __init__(self, fobj=sys.stdout):
        self.fobj = fobj

        self.display_names = {
            "fac_set": "Facilities",
        }

    def mk_fmt(self, *widths):
        return "%-" + "s %-".join(map(str, widths)) + "s"

    def mk_set_headers(self, data, columns):
        """ figure out sizes and create header fmt """
        columns = tuple(columns)
        lens = []

        for key in columns:
            value_len = max(len(str(each.get(key, ""))) for each in data)
            # account for header lengths
            lens.append(max(value_len, len(self._get_name(key))))

        fmt = self.mk_fmt(*lens)
        return fmt

    def _get_name(self, key):
        """ get display name for a key, or mangle for display """
        if key in self.display_names:
            return self.display_names[key]

        return key.capitalize()

    def _get_val(self, data, key):
        """ get value from a dict, format if necessary """
        return data.get(key, "")

    def _get_columns(self, data):
        """ get columns from a dict """
        return data.keys()

    def display_section(self, name):
        self._print(name)
        self._print("=" * len(name))
        self._print("")

    def display_headers(self, fmt, headers):
        self._print(fmt % headers)
        self._print(fmt % tuple("-" * len(x) for x in headers))

    def display_set(self, typ, data, columns):
        """ display a list of dicts """
        self.display_section("%s (%d)" % (self._get_name(typ), len(data)))
        headers = tuple(map(self._get_name, columns))
        fmt = self.mk_set_headers(data, columns)
        self.display_headers(fmt, headers)

        for each in data:
            row = tuple(self._get_val(each, k) for k, v in each.items())
            self._print(fmt % row)

        self._print("\n")

    def display_field(self, fmt, obj, field, display=None):
        if not display:
            display = self._get_name(field)
        self._print(fmt % (display, obj[field]))

    def check_set(self, data, name):
        if data.get(name, None):
            if hasattr(self, "print_" + name):
                getattr(self, "print_" + name)(data[name])

    def print_net(self, data):
        self.display_section("Network Information")
        fmt = "%-21s: %s"
        self.display_field(fmt, data, "name", "Name")
        self.display_field(fmt, data, "asn", "Primary ASN")
        self.display_field(fmt, data, "aka", "Also Known As")
        self.display_field(fmt, data, "website", "Website")
        self.display_field(fmt, data, "irr_as_set", "IRR AS-SET")
        self.display_field(fmt, data, "info_type", "Network Type")
        self.display_field(fmt, data, "info_prefixes6", "Approx IPv6 Prefixes")
        self.display_field(fmt, data, "info_prefixes4", "Approx IPv4 Prefixes")
        self.display_field(fmt, data, "looking_glass", "Looking Glass")
        self.display_field(fmt, data, "route_server", "Route Server")
        self.display_field(fmt, data, "created", "Created at")
        self.display_field(fmt, data, "updated", "Updated at")
        self._print("\n")

        self.display_section("Peering Policy Information")
        self.display_field(fmt, data, "policy_url", "URL")
        self.display_field(fmt, data, "policy_general", "General Policy")
        self.display_field(fmt, data, "policy_locations", "Location Requirement")
        self.display_field(fmt, data, "policy_ratio", "Ratio Requirement")
        self.display_field(fmt, data, "policy_contracts", "Contract Requirement")
        self._print("\n")

        self.check_set(data, "poc_set")
        self.check_set(data, "netixlan_set")
        self.check_set(data, "netfac_set")

    def print_poc_set(self, data):
        self.display_section("Contact Information")
        fmt = self.mk_fmt(6, 20, 15, 20, 14)
        hdr = ("Role", "Name", "Email", "URL", "Phone")
        self.display_headers(fmt, hdr)

        for poc in data:
            self._print(
                fmt
                % (
                    poc.get("role", ""),
                    poc.get("name", ""),
                    poc.get("email", ""),
                    poc.get("url", ""),
                    poc.get("phone", ""),
                )
            )

        self._print("\n")

    def print_netfac_set(self, data):
        self.display_section("Private Peering Facilities (%d)" % len(data))
        fmt = self.mk_fmt(51, 8, 15, 2)
        hdr = ("Facility Name", "ASN", "City", "CO")
        self.display_headers(fmt, hdr)
        for each in data:
            self._print(
                fmt
                % (
                    each.get("name", each.get("id")),
                    each.get("local_asn", ""),
                    each.get("city", ""),
                    each.get("country", ""),
                )
            )
        self._print("\n")

    def print_netixlan_set(self, data):
        self.display_section("Public Peering Points (%d)" % len(data))
        fmt = self.mk_fmt(36, 8, 27, 5)
        hdr = ("Exchange Point", "ASN", "IP Address", "Speed")
        self.display_headers(fmt, hdr)
        for ix in data:
            if ix.get("ipaddr4", None):
                self._print(
                    fmt
                    % (
                        ix.get("name", ix.get("ixlan_id")),
                        ix["asn"],
                        ix["ipaddr4"],
                        pretty_speed(ix["speed"]),
                    )
                )
            if ix.get("ipaddr6", None):
                if ix.get("ipaddr4", None):
                    self._print(fmt % ("", "", ix["ipaddr6"], ""))
                else:
                    self._print(
                        fmt
                        % (
                            ix["name"],
                            ix["asn"],
                            ix["ipaddr6"],
                            pretty_speed(ix["speed"]),
                        )
                    )
        self._print("\n")

    def _print(self, *args):
        """ internal print to self.fobj """
        string = " ".join(args) + "\n"
        self.fobj.write(string)

    def print(self, typ, data):
        """ *deprecated* - use display() """
        return self.display(typ, data)

    def display(self, typ, data):
        """ display section of typ with data """
        if hasattr(self, "print_" + typ):
            getattr(self, "print_" + typ)(data)

        elif not data:
            self._print("{}: {}".format(typ, data))

        elif isinstance(data, collections.Mapping):
            self._print("\n", typ)
            for k, v in data.items():
                self.print(k, v)

        elif isinstance(data, (list, tuple)):
            # tabular data layout for lists of dicts
            if isinstance(data[0], collections.Mapping):
                self.display_set(typ, data, self._get_columns(data[0]))
            else:
                for each in data:
                    self.print(typ, each)
        else:
            self._print("{}: {}".format(typ, data))

        self.fobj.flush()



from collections import OrderedDict

import munge.util
from twentyc.rpc import RestClient

import peeringdb
from peeringdb import config
from peeringdb import get_backend, resource
from peeringdb.resource import get_resource, Network
from peeringdb.sync import Updater, Fetcher


class _Query:
    "Wrapper to access a specific resource"

    def __init__(self, client, res):
        self.client = client
        self.res = res

    def get(self, pk):
        return self.client.get(self.res, pk)

    def all(self, **kwargs):
        return self.client.all(self.res, **kwargs)


class Client:
    "Main PeeringDB client."

    def __init__(self, cfg=None, **kwargs):
        """
        Arguments:
            - cfg <dict>: dict of complete config options (see config.ClientSchema),
                by default loads from DEFAULT_CONFIG_DIR
            - url<str>: URL to connect to
            - user<str>: username to connect to api with
            - password<str>: password
            - timeout<float>: timeout to fail after
        """
        if cfg is None:
            cfg = config.load_config()
        self.config = cfg
        orm_config = cfg["orm"]
        orm_name = orm_config["backend"]
        if not peeringdb.backend_initialized():
            peeringdb.initialize_backend(orm_name, **orm_config)

        sync_config = cfg["sync"]
        # override config with kwargs
        munge.util.recursive_update(sync_config, kwargs)

        self._fetcher = Fetcher(**sync_config)
        self._updater = Updater(self._fetcher, **sync_config)

        self.update_all = self._updater.update_all
        self.update = self._updater.update
        self.update_where = self._updater.update_where

        tag_res = OrderedDict(
            [(res.tag, _Query(self, res)) for res in resource.all_resources()]
        )
        tag_attrs = {
            **tag_res,
            **{
                "keys": lambda self: list(tag_res.keys()),
                "all": lambda self: list(tag_res.values()),
            },
        }
        self._Tags = type("_Tags", (), tag_attrs)
        self.tags = self._Tags()

    def fetch(self, R, pk, depth=1):
        "Request object from API"
        d, e = self._fetcher.fetch(R, pk, depth)
        if e:
            raise e
        return d

    def fetch_all(self, R, depth=1, **kwargs):
        "Request multiple objects from API"
        d, e = self._fetcher.fetch_all(R, depth, kwargs)
        if e:
            raise e
        return d

    def get(self, res, pk):
        "Get a resource instance by primary key (id)"
        B = get_backend()
        return B.get_object(B.get_concrete(res), pk)

    def all(self, res):
        "Get resources using a filter condition"
        B = get_backend()
        return B.get_objects(B.get_concrete(res))

    @property
    def backend(self):
        return get_backend()



"""
Sync implementation module
"""
import logging
import datetime
from collections import defaultdict
from peeringdb.resource import get_resource
from peeringdb import get_backend
from peeringdb.util import group_fields


def _field_resource(B, concrete, field):
    return B.get_resource(B.get_field_concrete(concrete, field))


def clean_helper(B, obj, clean_func):
    """
    Clean object, intercepting and collecting any missing-relation or
    unique-constraint errors and returning the relevant resource ids/fields.

    Returns:
        - tuple: (<dict of non-unique fields>, <dict of missing refs>)
    """
    try:
        clean_func(obj)
    except B.validation_error() as e:
        # _debug.log_validation_errors(B, e, obj, k)

        # Check if it's a uniqueness or missing relation error
        fields = B.detect_uniqueness_error(e)
        missing = B.detect_missing_relations(obj, e)
        return fields, missing
    return ({}, {})


def initialize_object(B, res, row):
    """
    Do a shallow initialization of an object

    Arguments:
        - row<dict>: dict of data like depth=1, i.e. many_refs are only ids
    """
    field_groups = group_fields(B.get_concrete(res))

    try:
        obj = B.get_object(B.get_concrete(res), row["id"])
    except B.object_missing_error(B.get_concrete(res)):
        tbl = B.get_concrete(res)
        obj = tbl()
    return obj


def set_scalars(B, res, obj, row):
    field_groups = group_fields(B.get_concrete(res))
    # Set attributes, refs
    for fname, field in field_groups["scalars"].items():
        value = row.get(fname, getattr(obj, fname, None))
        value = B.convert_field(obj.__class__, fname, value)
        setattr(obj, fname, value)
    # _debug('res, row: %s, %s', res, row)


def _get_subrow(row, fname, field):
    key = field.column
    subrow = row.get(key)
    if subrow is None:  # e.g. use "org" if "org_id" is missing
        key = fname
        subrow = row[key]
    return key, subrow


def extract_relations(B, res, row):
    field_groups = group_fields(B.get_concrete(res))
    # Already-fetched, and id-only refs
    fetched, dangling = defaultdict(dict), defaultdict(set)

    # Handle subrows that might be shallow (id) or deep (dict)
    def _handle_subrow(R, subrow):
        if isinstance(subrow, dict):
            pk = subrow["id"]
            fetched[R][pk] = subrow
        else:
            pk = subrow
            dangling[R].add(pk)
        return pk

    for fname, field in field_groups["single_refs"].items():
        fieldres = _field_resource(B, B.get_concrete(res), fname)
        _, subrow = _get_subrow(row, fname, field)
        pk = _handle_subrow(fieldres, subrow)

    for fname, field in field_groups["many_refs"].items():
        fieldres = _field_resource(B, B.get_concrete(res), fname)
        for subrow in row.get(fname, []):
            _handle_subrow(fieldres, subrow)

    return fetched, dangling


def set_single_relations(B, res, obj, row):
    field_groups = group_fields(B.get_concrete(res))
    for fname, field in field_groups["single_refs"].items():
        key, subrow = _get_subrow(row, fname, field)
        if isinstance(subrow, dict):
            pk = subrow["id"]
        else:
            pk = subrow
        setattr(obj, key, pk)


def set_many_relations(B, res, obj, row):
    field_groups = group_fields(B.get_concrete(res))
    for fname, field in field_groups["many_refs"].items():
        fieldres = _field_resource(B, B.get_concrete(res), fname)
        pks = row.get(fname, [])
        objs = [B.get_object(B.get_concrete(fieldres), pk) for pk in pks]
        B.set_relation_many_to_many(obj, fname, objs)


def patch_object(B, res, obj, strip_tz):
    field_groups = group_fields(B.get_concrete(res))

    for fname in field_groups["scalars"]:
        value = getattr(obj, fname)
        if strip_tz and isinstance(value, datetime.datetime):
            value = value.replace(tzinfo=None)
        B.update(obj, fname, value)



"""
Wrapper module for sequential task (continuation) handling.
Use as fallback for _tasks_async when async/await is not available.
"""
from types import GeneratorType


class UpdateTask:
    "Wrap a generator in a task-like interface"

    def __init__(self, gen, desc):
        """
        Arguments:
            - gen: generator object
            - desc<tuple>: (Resource, key)
        """
        assert isinstance(gen, GeneratorType)
        self._gen = gen
        self._desc = desc

    def __repr__(self):
        res, pk = self._desc
        return "<UpdateTask: ({}, {})>".format(res.tag, pk)

    def cancel(self):
        pass

    def __iter__(self):
        return self._gen

    def send(self, x):
        return self._gen.send(x)


def gather(jobs):
    "Aggregate and collect jobs"
    for job in jobs:
        yield from job


def wrap_generator(func):
    "Identity decorator (only needed for async compatibility)"
    return func


def _consume_task_or_generator(item):
    if isinstance(item, (GeneratorType, UpdateTask)):
        return _consume_task(item)
    else:
        return item


def _consume_task(gen):
    r, ret = None, []
    while True:
        try:
            item = gen.send(r)
        except StopIteration:
            break
        r = _consume_task_or_generator(item)
        ret.append(r)

    if len(ret) == 1:
        return ret.pop()
    return ret


def run_task(func):
    """
    Decorator to collect and return generator results, returning a list
    if there are multiple results
    """

    def _wrapped(*a, **k):
        gen = func(*a, **k)
        return _consume_task(gen)

    return _wrapped



from setuptools import setup, find_packages

version = open("facsimile/VERSION").read().strip()
requirements = open("facsimile/requirements.txt").read().split("\n")
test_requirements = open("facsimile/requirements-test.txt").read().split("\n")

setup(
    name="peeringdb",
    version=version,
    author="PeeringDB",
    author_email="support@peeringdb.com",
    description="peeringdb client and interface",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Telecommunications Industry",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Internet",
    ],
    packages=["peeringdb"],
    url="https://github.com/peeringdb/peeringdb-py",
    download_url="https://github.com/peeringdb/peeringdb-py/%s" % version,
    include_package_data=True,
    install_requires=requirements,
    test_requires=test_requirements,
    entry_points={"console_scripts": ["peeringdb=peeringdb.cli:main",]},
    zip_safe=True,
)



"""
PeeringDB resource definitions
"""
from collections import OrderedDict

# Generate classes
_NAMES = OrderedDict(
    [
        ("org", "Organization"),
        ("fac", "Facility"),
        ("net", "Network"),
        ("ix", "InternetExchange"),
        ("ixfac", "InternetExchangeFacility"),
        ("ixlan", "InternetExchangeLan"),
        ("ixpfx", "InternetExchangeLanPrefix"),
        ("netfac", "NetworkFacility"),
        ("netixlan", "NetworkIXLan"),
        ("poc", "NetworkContact"),
    ]
)

RESOURCES_BY_TAG = OrderedDict()

for tag, name in _NAMES.items():

    class Meta(type):
        def __repr__(cls, _name=name):
            return _name

    Class = Meta(name, (), {"tag": tag})

    RESOURCES_BY_TAG[tag] = Class
    locals()[name] = Class

is_resource_tag = RESOURCES_BY_TAG.__contains__


def get_resource(tag):
    return RESOURCES_BY_TAG[tag]


get = get_resource


def all_resources():
    return list(RESOURCES_BY_TAG.values())



from functools import wraps
import inspect

from peeringdb.resource import RESOURCES_BY_TAG


def reftag_to_cls(fn):
    """
    decorator that checks function arguments for `concrete` and `resource`
    and will properly set them to class references if a string (reftag) is
    passed as the value
    """
    spec = inspect.getfullargspec(fn)
    names, values = spec.args, spec.defaults

    @wraps(fn)
    def wrapped(*args, **kwargs):
        i = 0
        backend = args[0]
        for name in names[1:]:
            value = args[i]
            if name == "concrete" and isinstance(value, str):
                args[i] = backend.REFTAG_CONCRETE[value]
            elif name == "resource" and isinstance(value, str):
                args[i] = backend.REFTAG_RESOURCE[value]
            i += 1
        return fn(*args, **kwargs)

    return wrapped


class Field:
    """
    We use this to provide field instances to backends that
    don't use classes to describe their fields
    """

    def __init__(self, name):
        self.name = name
        self.column = column


class EmptyContext:

    """
    We use this to provide a dummy context wherever it's optional
    """

    def __enter__(self):
        pass

    def __exit__(self, *args):
        pass


class Base:

    """
    Backend base class.

    Do NOT extend this directly when implementing a new backend, instead
    extend Interface below.
    """

    # Handleref tag to resource class mapping
    REFTAG_RESOURCE = RESOURCES_BY_TAG

    @property
    def CONCRETE_MAP(self):
        if not hasattr(self, "_CONCRETE_MAP"):
            self._CONCRETE_MAP = {
                concrete: res for (res, concrete) in self.RESOURCE_MAP.items()
            }
        return self._CONCRETE_MAP

    def get_concrete(self, res):
        """
        returns the concrete class for the resource

        Returns:

            - concrete class
        """
        return self.RESOURCE_MAP[res]

    def is_concrete(self, cls):
        """
        check if concrete class exists in the resource -> concrete mapping

        Returns:

            - bool: True if class exists in the resource -> concrete mapping
        """
        return cls in self.CONCRETE_MAP

    def get_resource(self, cls):
        """
        returns the resource class for the concrete class

        Returns:

            - resource class
        """
        return self.CONCRETE_MAP[cls]


class Interface(Base):

    """
    backend adapter interface

    extend this when making a new backend
    """

    # Resource class to concrete class mapping
    # should go in here
    RESOURCE_MAP = {}

    # Handleref tag to concrete class mapping
    # should go in here
    REFTAG_CONCRETE = {}

    @classmethod
    def validation_error(cls, concrete=None):
        """
        should return the exception class that will
        be raised when an object fails validation

        Arguments:

            - concrete: if your backend has class specific validation
                errors and this is set, return the exception class that would be
                raised for this concrete class.

        Returns:

            - Exception class
        """
        return Exception

    @classmethod
    def object_missing_error(cls, concrete=None):
        """
        should return the exception class that will
        be raised when an object cannot be found

        Arguments:

            - concrete: if your backend has class specific object missing
                errors and this is set, return the exception class that would be
                raised for this concrete class.

        Returns:

            - Exception class
        """
        return Exception

    @classmethod
    def atomic_transaction(cls):
        """
        Allows you to return an atomic transaction context
        if your backend supports it, if it does not, leave as is

        This should never return None

        Returns:

            - python context instance
        """
        return EmptyContext()

    @classmethod
    def setup(cls):
        """
        operations that need to be done ONCE during runtime
        to prepare usage for the backend
        """
        pass

    # INTERFACE (REQUIRED)
    # The following methods are required to be overwritten in
    # your backend and will raise a NotImplementedError if
    # they are not.
    #
    # when overriding make sure you also apply the `reftag_to_cls`
    # decorator on the methods that need it

    @reftag_to_cls
    def create_object(self, concrete, **data):
        """
        should create object from dict and return it

        Arguments:

            - concrete: concrete class

        Keyword Arguments:

            - object field names -> values
        """
        raise NotImplementedError()

    # TODO:
    def delete_all(self):
        """
        Delete all objects, essentially empty the database
        """
        raise NotImplementedError()

    def detect_missing_relations(self, obj, exc):
        """
        Should parse error messages and collect the missing relationship
        errors as a dict of Resource -> {id set} and return it

        Arguments:

            - obj: concrete object instance
            - exc: exception instance

        Returns:

            - dict: {Resource : [ids]}
        """
        raise NotImplementedError()

    def detect_uniqueness_error(self, exc):
        """
        Should parse error message and collect any that describe violations
        of a uniqueness constraint.

        return the curresponding fields, else None

        Arguments:

            - exc: exception instance

        Returns:

            - list: list of fields
            - None: if no uniqueness errors
        """
        raise NotImplementedError()

    @reftag_to_cls
    def get_field_names(self, concrete):
        """
        Should return a list of field names for the concrete class

        Arguments:

            - concrete: concrete class

        Returns:

            - list: [<str>,...]
        """
        raise NotImplementedError()

    @reftag_to_cls
    def get_field_concrete(self, concrete, field_name):
        """
        Return concrete class for relationship by field name

        Arguments:

            - concrete: concrete class
            - field_name

        Returns:

            - concrete class
        """
        raise NotImplementedError()

    @reftag_to_cls
    def get_object(self, concrete, id):
        """
        should return instance of object with matching id

        Arguments:

            - concrete: concrete class
            - id: object primary key value

        Returns:

            - concrete instance
        """
        raise NotImplementedError()

    @reftag_to_cls
    def get_object_by(self, concrete, field_name, value):
        """
        very simply search function that should return
        collection of objects where field matches value

        Arguments:

            - concrete: concrete class
            - field_name: query this field for a match
            - value: match this value (simple equal matching)

        Returns:

            - concrete instance
        """
        raise NotImplementedError()

    @reftag_to_cls
    def get_objects(self, concrete, ids=None):
        """
        should return collection of objects

        Arguments:

            - concrete: concrete class
            - ids: if specified should be a list of primary
                key values and only objects matching those
                values should be returned

        Returns:

            - collection of concrete instances
        """
        raise NotImplementedError()

    @reftag_to_cls
    def get_objects_by(self, concrete, field, value):
        """
        very simple search function that should return
        collection of objects where field matches value

        Arguments:

            - concrete: concrete class
            - field_name: query this field for a match
            - value: match this value (simple equal matching)

        Returns:

            - collection of concrete instances
        """
        raise NotImplementedError()

    @reftag_to_cls
    def is_field_related(self, concrete, field_name):
        """
        Should return a tuple containing bools on whether
        a field signifies a relationship and if it's a single
        relationship or a relationship to multiple objects

        Arguments:

            - concrete: concrete class
            - field_name: query this field for a match

        Returns:

            - tuple: (bool related, bool many)
        """
        raise NotImplementedError()

    @reftag_to_cls
    def last_change(self, concrete):
        """
        should return unix epoch timestamp of the `updated` field
        of the most recently updated object

        Arguments:

            - concrete: concrete class

        Returns:

            - int
        """
        raise NotImplementedError()

    def save(self, obj):
        """
        Save the object instance

        Arguments:

            - obj: concrete object instance
        """
        raise NotImplementedError()

    def set_relation_many_to_many(self, obj, field_name, objs):
        """
        Setup a many to many relationship

        Arguments:

            - obj: concrete object instance
            - field_name: name of the field that holds the relationship
            - objs: collection of concrete objects to setup relationships with
        """
        raise NotImplementedError()

    def update(self, obj, field_name, value):
        """
        update field on a concrete instance to value

        this does not have to commit to the database, which will be
        handled separately via the `save` method.

        Arguments:

            - obj: concrete object instance
            - field_name
            - value
        """
        setattr(obj, field_name, value)

    ## INTERFACE (OPTIONAL / SITUATIONAL)

    @reftag_to_cls
    def get_field(self, concrete, field_name):
        """
        Should retrun a field instance, if your backend does not use
        classes to describe fields, leave this as is

        Arguments:

            - concrete: concrete class
            - field_name

        Returns:

            - field instance
        """
        return Field(field_name)

    @reftag_to_cls
    def get_fields(self, concrete):
        """
        Should return a collection of fields, if your backend does not
        use classes to describe fields, leave this as is

        Arguments:

            - concrete: concrete class

        Returns:

            - collection of field instances
        """
        return [Field(name) for name in self.field_names(concrete)]

    def clean(self, obj):
        """
        Should take an object instance and validate / clean it

        Arguments:

            - obj: concrete object instance
        """
        pass

    @reftag_to_cls
    def convert_field(self, concrete, field_name, value):
        """
        Should take a value and a field definition and do a value
        conversion if needed.

        should return the new value.

        Arguments:

            - concrete: concrete class
            - field_name
            - value
        """
        pass

    def migrate_database(self, verbosity=0):
        """
        Do database migrations

        Arguments:

            - verbosity <int>: arbitrary verbosity setting, 0 = be silent,
                1 = show some info about migrations.
        """
        pass

    def is_database_migrated(self, **kwargs):
        """
        Should return whether the database is fully migrated

        Returns:

            - bool
        """
        return True
